#include "Util.h"
#include "EnergyModel.h"
#include "Input.h"

// ***************************************************************************************************************************************************************************************************
// MCMC based analysis of SULDEX data 
// (Inference of disassociation constants from high-throughput massively parallel binding/sequencing experiments)
// ***************************************************************************************************************************************************************************************************

// Original version programmed by Hyunmin Kim
// Extensively modified and expanded by A.P. Jason de Koning (2010) 
// jason.de.koning@gmail.com  http://jasondk.org

// David D. Pollock lab, University of Colorado Denver School of Medicine
// http://www.EvolutionaryGenomics.com


// ***** !!IMPORANT!! ********************************************************************************************************************************************************************************
// ** TO USE, THE PROPOSALS MUST BE MANUALLY TUNED TO GIVE A REASONABLE ACCEPTANCE RATE.  FAILURE TO DO THIS (especially for Log([TF])) WILL LEAD TO POOR RESULTS.
// ** (the asymptotically optimal acceptance rate is ~45% for unidimensional proposal)
// ***************************************************************************************************************************************************************************************************

// Jason's change log:

// - changed most of the proposal distributions from uniform to gaussian to aid with mixing
// - modified to allow oligos with zero counts in one fraction or the other (pre or post)
// - added an updater on the prebound frequencies, {W}
// - fixed bugs relating to input of reference Kd's (original program assumed input Kd's were integers, causing problems with real data)

// - implemented a set of partial likelihood function evaluations to speedup evaluation of proposals affecting only parts of the overall likelihood fxn (yields much faster updaters)
// - implemented work arounds for the clunky Checkpoint() / Rollback() routines that were backing up and restoring all parameters with every proposal
//		-- disabled copying the sequence data over and over with these routines (was unnecessary and time consuming)
//		-- replaced general usage with parameter specific backup/restore (much faster)
// - implemented a lookup table for logs using precomputed log's and linear interpolation (except for LogFactorial function, which is only called to get the full likelihood anyway -i.e., not used very often)
// - blocked the conditional on the RJ move for deleting a term: will now evaluate even if energy parameter is larger than threshold (not needed when using proposal reallocation, which we are)

// - because log([TF]) can be non-identifiable with the sum over all delta G's with no reference oligos (and nearly so with only a few reference oligos), modified the proposal mechanism for logTF to scale deltaGs with logTF
//		-- mixes much better now; much faster
// - added an updater to scale all the energy parameters at once ("Esc") - makes mixing much better for energy model, and greatly reduced the variance of chains on test data
// - removed updater on Eopt - now required a priori

// - CONTROL FILE (MCMC_CONTROL.txt): set bounds on log-based parameters to symmetric - range: negative to positive (not 0 to x)

// * NOTE: because of partial likelihood function evaluations, can no longer count on _llh as being the current
// *       likelihood.  Need to do an extra CalcLlh() call before outputting!

// ***************************************************************************************************************************************************************************************************

int main( int argc, char *argv[]) {
	if(argc < 3){
		cerr<<"usage: "<<argv[0]<<" <sdx input> <method> [<ctr>]"<<endl;
		return(0);
	}
	string buf = "";
	for(int i=1; i< argc; i++) buf += " " + string(argv[i]);
	
	string fin_sdx;
	string method;
	istringstream iss(buf);iss>>fin_sdx>>method;
	string fcon_out = fin_sdx.substr(0,fin_sdx.size()-4)+"_"+method+".ctrl";
	string fcon = "MCMC_CONTROL.txt";
	if(!iss.eof()){
		iss>>fcon;
	}else if (fexist(fcon_out.c_str())){
		fcon = fcon_out;
		
	}
	cerr<<"use control: "<<fcon<<endl;
	
	Parameter param(fin_sdx);
	Input ip(fcon);
	ofstream ofs(fcon_out.c_str());
	ofs<<ip;
	ofs.close();
	
	srand(time(0));
	EModel em(&ip,&param);
	cerr<<em;
	
	string fout = fin_sdx.substr(0,fin_sdx.size()-4)+"_"+method+".mcmc";
	srand(time(0));
	em.MCMC(fout,method);
	cerr<<"ok"<<endl;
	return(0);
}
