/*
 *
 *  Created by Hyunmin Kim.
 *
 */

#ifndef _EnergyModel_H_
#define _EnergyModel_H_

#include <map>
#include <vector>
#include <algorithm>
#include <fstream>
#include "Util.h"
#include "Input.h"

struct PropWind{
	PropWind(double w, double min, double max){
		_w =w; _min=min; _max=max;
	}
	double _w; //window
	double _min; //lower bound
	double _max; //higher bound
	double _v; // previous values
	double Propose( double prev) { 
		if(_w == 0) return(prev);
		_v = prev;
		double change = ( _rng.normal( 0.0, (_w / 2.0 ) ) );
		double proposal = prev + change;
		if( proposal > _max || proposal < _min)
			proposal = prev - change;
		return( proposal);
	}
	
	double GetPrev(){ return(_v);}
};

class Parameter{
private:
	int 	_l;	// size of sequence
	int 	_n; 	// number of sequences
	double 	_tf; // original tf
	double 	_lntf; 	// log [TF]
	double  _RT; 	// Kcal/(K*mol)
	
	double 	*_E; 	// additive energy model
	//vector<double>	_X;	// number of values
	int	*_D;	// Delta matrix
	double 	_Eopt;	// optimum binding energy
	
	int	*_S; 	// sequence
	bool 	*_R; 	// indicator for reference sequences
	double 	*_W; 	// freq. of prior to binding
	int 	*_C;	// count before binding
	int	_Ctot;
	int 	*_B; 	// count after binding
	int 	_Btot; 	// total binding
	double 	*_P; 	// binding probabilities
	double 	_Pave; 	// average binding probability
	double 	*_F; 	// frequency after biding
	double 	*_G; 	// binding energy
	double 	*_GE; 	// binding energy  estimated by additive model
	double _sigE;  // energy error 
	
public:
	Parameter(int n, int l);
	Parameter(Parameter&);
	Parameter(string&);
	~Parameter();
	
	bool CheckInput(string&);
	void Read(string&);
	
	void InitDeltaGs(void);
	
	double deltaG(double nanomole){ return(_RT *(log(nanomole)-log(1000000000))); }
	double nanoKd(double deltag){ return(exp(deltag/_RT+log(1000000000))); }
	int n(){return(_n);}
	int l(){return(_l);}
	double E(int i, int j){ return(_E[i*4+j]);}
	int D(int i, int j){ return(_D[i*4+j]);}
	//double X(int i){ return(_X[i]);}
	//int K(){return(_X.size());}
	double Eopt(){ return(_Eopt);}
	
	int S(int n, int l){ return(_S[n*_l+l]);}
	bool R(int n){ return(_R[n]);}
	int B(int n){ return(_B[n]);}
	int Btot(){ return(_Btot);}
	int C(int n){ return(_C[n]);}
	int Ctot(){ return(_Ctot);}
	double W(int n){ return(_W[n]); }
	double P(int n){ return(_P[n]);}
	double Pave(){ return(_Pave);}
	double F(int n){ return(_F[n]);}
	double G(int n){ return(_G[n]);}
	double GE(int n){ return(_GE[n]);}
	double sigE(){return(_sigE);}
	
	double lntf(){ return(_lntf);}
	
	void SetEopt(double v){ _Eopt = v;}
	void SetE(int i, int j, double v){ _E[i*4+j] = v;}
	void SetD(int i, int j, int v){ _D[i*4+j] = v;}
	
	void SetS(int n, int l,int v){ _S[n*_l+l] = v;}
	void SetR(int n, bool v){ _R[n]=v;}
	void SetB(int n, int v){ _B[n]=v;}
	void SetC(int n, int v){ _C[n]=v;}
	void SetBtot(int v){ _Btot=v;}
	void SetCtot(int v){ _Ctot=v;}
	void SetW(int n, double v){ _W[n] = v;}
	void SetP(int n, double v){ _P[n]=v;}
	void SetPave(double v){ _Pave = v; }
	void SetF(int n, double v){ _F[n]=v;}
	void SetG(int n, double v){ _G[n]=v;}
	void SetGE(int n, double v){ _GE[n]=v;}
	void SetLntf(double v){ _lntf = v;}
	void SetSigE(double v){ _sigE = v;}
	
	void Copy(Parameter &);
	void Init(int n, int l);
	void Destroy();
	double CalcDeltaE(int);
	void UpdateP(int m){ 
		_P[m] = 1/(1+exp(_G[m]-_lntf)); 
	}
	
	void UpdateGE(int m){ _GE[m] = _Eopt + CalcDeltaE(m); }
	double Constrain(int i, int j, double v); 
	void NormalizeW(){
		// W[k] is the freq prior to binding
		double _Wsum = 0;
		for(int k=0; k<_n; k++) _Wsum += _W[k];
		for(int k=0; k<_n; k++) _W[k] = _W[k]/_Wsum;
	}
	void Normalize(){
		double minVal = 0.0000000000001;
		
		_Pave = 0;
		for(int k=0; k<_n; k++) {
			if ( _W[k]*_P[k] <= minVal ) {
				_Pave += minVal;
			} else {
				_Pave += _W[k]*_P[k];
			}
		}
		for(int k=0; k<_n; k++) {
			if ( _W[k] * _P[k] <= 0.0 ) {
				_F[k] = minVal;
			} else {
				_F[k] = _W[k]*_P[k]/_Pave;
			}
		}
	}
	bool Birth(int i, int j, double u);
	bool Death(int i, int j, double u);
	
	double CalcLlh();
	double CalcLlh_wrtFreqsOnly();
	double CalcLlh_wrtPostboundFreqsOnly();
	double CalcLlh_wrtEnergyModelOnly();
	
	double CalcLlh(int k);
	int CntS(int i,int j){
		int cnt = 0;
		for(int m=0; m< n(); m++) if(S(m,i)==j) cnt++;
		return(cnt);
	}
	int CntD(int i){
		int cnt = 0;
		for(int j=0; j<4; j++) if(D(i,j) > cnt) cnt = D(i,j);
		return(cnt);
	}
	int CntD(int i, int k){
		int cnt = 0;
		for(int j=0; j<4; j++) if(D(i,j)==k) cnt++;
		return(cnt);
	}
	int FndD(int i, int k, int from=0){
		for(int j=from; j<4; j++)
			if(D(i,j)==k) return(j);
		return(-1);
	}
	bool UpdateModelZeroOne(int,int,double);
	friend ostream & operator << (ostream&, Parameter &);
};
class EModel{
private: 
	bool _fixdmat;
	Parameter * _param;
	Parameter * _param_c;
	Parameter * _param_c2;
	
	double _llh;
	double _maxLlh;
	Input * _ctrl;
	int _accG; int _tryG; int _tryW; int _accW;
	int _accE; int _tryE;		int _accEscale; int _tryEscale;
	int _accLntf; int _tryLntf;
	int _accEopt; int _tryEopt;
	int _accSigE; int _trySigE;
	int _accJ; int _tryJ;
	
	PropWind * _pwG;
	PropWind * _pwW;
	PropWind * _pwE;
	PropWind * _pwLntf;
	PropWind * _pwEopt;
	PropWind * _pwSigE;
	
	vector< vector<double> > _Epost;
	
	
public:
	EModel(Input*, Parameter*);
	~EModel();
	
	void Destroy();
	void UpdateModel(int i, int j, double jp);
	void WriteParam(ostream &, Input &, bool header=false);
	void Read(string &);
	void Clear();
	void MCMC(string&, string&);
	friend ostream & operator << (ostream&, EModel&);
	double CalcLlh();
	bool Accept(double);
	void UpdateModelZeroOne(bool error);
	bool UpdateE(int,int);
	bool UpdateE();
	bool UpdateEscale();
	bool UpdateG(int iter);
	bool UpdatePreBoundFreqs();
	bool UpdateSigE();
	bool UpdateEopt();
	bool UpdateEoptE(double);
	bool UpdateLntf();
	
	void CheckPoint1(){ _param_c->Copy( *_param);}
	void Rollback1(){ _param->Copy( *_param_c);}
	void CheckPoint2(){ _param_c2->Copy(*_param);}
	void Rollback2(){ _param->Copy( *_param_c2);}
};

#endif 
